// ----------------------------------------------------------------------------
/// gpio.h
/// © 2009-2012 Sealevel Systems, Inc.
/// <summary>
/// Some helper functions to aid in dealing with sysfs gpio.
/// </summary>
///
/// <remarks>
/// Both maps for the R9 gpio layout as well as some functions that will
/// register, read, write, and release gpio are contained in this header file.
/// </remarks>
// ----------------------------------------------------------------------------

#ifndef __SEALEVEL_GPIO_H__
#define __SEALEVEL_GPIO_H__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

// ARM GPIO offsets
#define ARM_GPIO_PA 	0
#define ARM_GPIO_PB	32
#define ARM_GPIO_PC	64
#define ARM_GPIO_PD	96
#define ARM_GPIO_PE	128

// Sysfs gpio driver interface path
#define SYSFS_GPIO	"/sys/class/gpio"

// ----------------------------------------------------------------------------
// Helper functions, to make accessing the sysfs gpio easier
// ----------------------------------------------------------------------------

// This is a check to see if a gpio point is already registered in sysfs
int gpio_is_open(int offset)
{
	// to check wether an ARM gpio point has been registered (exported),
	// all we need to do is look for the directory /sys/class/gpio/gpio#
	struct stat buf;
	char fileName[256];

	if (offset < 0)
	{
		return 0;
	}

	sprintf(fileName, "%s/gpio%d", SYSFS_GPIO, offset);
	return (0 == stat(fileName, &buf));
}

// This is a simple function to write a string to the file specified
int write_file(char *fileName, char *string)
{
	int ret;
	FILE *file = fopen(fileName, "w");
	if (file == NULL)
	{
		return -1;
	}

	ret = fputs(string, file);
	fclose(file);

	return ret;
}

// This simple function will read a string (up to maxSize) from the file specified
int read_file(char *fileName, char *string, int maxSize)
{
	FILE *file = fopen(fileName, "r");
	if (file == NULL)
	{
		return -1;
	}

	if (fgets(string, maxSize, file) == NULL)
	{
		fclose(file);
		return -2;
	}
	
	fclose(file);
	return 0;
}

int gpio_register(int offset)
{
	if (offset < 0)
	{
		return -1;
	}

	if (!gpio_is_open(offset))
	{
		char fileName[256];
		char command[256];

		sprintf(fileName, "%s/export", SYSFS_GPIO);
		sprintf(command, "%d", offset);

		if (write_file(fileName, command) < 0)
		{
			return -2;
		}
		
		if (!gpio_is_open(offset))
		{
			return -3;
		}
	}

	return 0;
}

void gpio_unregister(int offset)
{
	if (offset < 0)
	{
		return;
	}

	if (gpio_is_open(offset))
	{
		char fileName[256];
		char command[256];

		sprintf(fileName, "%s/unexport", SYSFS_GPIO);
		sprintf(command, "%d", offset);

		write_file(fileName, command);
	}
}

int gpio_set_direction(int offset, char *direction)
{
	char fileName[256];
	char temp[256];

	if (offset < 0)
	{
		return -1;
	}

	if (!gpio_is_open(offset))
	{
		return -2;
	}

	sprintf(fileName, "%s/gpio%d/direction", SYSFS_GPIO, offset);
	
	if (read_file(fileName, temp, sizeof(temp)) < 0)
	{
		return -3;
	}

	if (strncmp(temp, direction, strlen(direction)) == 0)
	{
		return 0;
	}

	if (write_file(fileName, direction) < 0)
	{
		return -4;
	}

	return 0;
}

int gpio_set_value(int offset, int value)
{
	char fileName[256];
	char command[256];

	if (offset < 0)
	{
		return -1;
	}

	if (!gpio_is_open(offset))
	{
		return -2;
	}

	sprintf(fileName, "%s/gpio%d/value", SYSFS_GPIO, offset);
	sprintf(command, "%s", value ? "1" : "0");

	if (write_file(fileName, command) < 0)
	{
		return -3;
	}

	return 0;
}

int gpio_get_value(int offset, int *value)
{
	char fileName[256];
	char command[256];

	if (offset < 0)
	{
		return -1;
	}

	if (!gpio_is_open(offset))
	{
		return -2;
	}

	sprintf(fileName, "%s/gpio%d/value", SYSFS_GPIO, offset);

	if (read_file(fileName, command, sizeof(command)) < 0)
	{
		return -3;
	}

	if (strncmp("0", command, 1) == 0)
	{
		*value = 0;
	}
	else if (strncmp("1", command, 1) == 0)
	{
		*value = 1;
	}
	else
	{
		return -4;
	}

	return 0;	
}

#endif
