// ----------------------------------------------------------------------------
/// ls7366.h
/// © 2009-2012 Sealevel Systems, Inc.
/// <summary>
/// LSI/CSI 32-bit Quadrature Counter with SPI Interface protocol definitions.
/// </summary>
///
/// <remarks>
/// This header defines several constants useful for communicating with the
/// LSI/CSI LS7366 SPI attached 32-bit Quadrature counter.  There are several
/// modes of operation for this counter, and full documentation is available
/// in the LS7366 datasheet 7366-083005-1.
/// </remarks>
// ----------------------------------------------------------------------------

// 8-bit Instruction Register
#define INSTRUCTION_MASK 	~0xF8
// D7 - D6:	Instruction function selection.
//		= 00 	Clear register
//		= 01	Read register
//		= 10	Write register
//		= 11	Load register
#define INSTRUCTION_CMD_CLEAR	0x00
#define INSTRUCTION_CMD_READ	0x40
#define INSTRUCTION_CMD_WRITE	0x80
#define INSTRUCTION_CMD_LOAD	0xC0
#define INSTRUCTION_CMD_MASK	~0xC0
// D5 - D3:	Instruciton register selection.
//		= 000	None
//		= 001	MDR0
//		= 010	MDR1
//		= 011	DTR
//		= 100	CNTR
//		= 101	OTR
//		= 110	STR
//		= 111	None
#define INSTRUCTION_REG_NONE	0x00
#define INSTRUCTION_REG_MDR0	0x08
#define INSTRUCTION_REG_MDR1	0x10
#define INSTRUCTION_REG_DTR	0x18
#define INSTRUCTION_REG_CNTR	0x20
#define INSTRUCTION_REG_OTR	0x28
#define INSTRUCTION_REG_STR	0x30
#define INSTRUCTION_REG_MASK	~0x38
// D2 - D0:	Instruction unused.
//		= XXX
#define INSTRUCTION_UNUSED_MASK ~0x07


// 8-bit Mode Register 0
#define MDR0_MASK		~0xFF
// D7:		Filter Clock Division Factor.
// 		= 0 	Factor = 1
// 		= 1	Factor = 2
#define MDR0_FILTER_FACTOR_1	0x00
#define MDR0_FILTER_FACTOR_2	0x80
#define MDR0_FILTER_FACTOR_MASK	~0x10
// D6:		Index input polarity.  (This option doesn't seem to work...)
// 		= 0 	Negative input
//		= 1	Positive input
#define MDR0_INDEX_NEGATIVE	0x00
#define MDR0_INDEX_POSITIVE	0x40
#define MDR0_INDEX_MASK		~0x40
// D5 - D4:	Index mode.
//		= 00	Disabled
//		= 01	Load CNTR (transfers DTR to CNTR)
//		= 10	Reset CNTR (sets CNTR to 0)
//		= 11	Load OTR (transfers CNTR to OTR)
#define MDR0_INDEX_DISABLED	0x00
#define MDR0_INDEX_LOAD_CNTR	0x10
#define MDR0_INDEX_RESET_CNTR	0x20
#define MDR0_INDEX_LOAD_OTR	0x30
#define MDR0_INDEX_MODE_MASK	~0x30
// D3 - D2:	Counter mode.
//		= 00	Free-running
//		= 01	Single-cycle count
//		= 10	Range-limit count
//		= 11	Modulo-n count
#define MDR0_MODE_FREE		0x00
#define MDR0_MODE_SINGLE	0x04
#define MDR0_MODE_RANGE		0x08
#define MDR0_MODE_MODULO	0x0C
#define MDR0_MODE_MASK		~0x0C
// D1 - D0: 	Quadrature mode.
#define MDR0_NONQUAD		0x00
#define MDR0_QUAD		0x01
#define MDR0_QUAD_2X		0x02
#define MDR0_QUAD_4X		0x03
#define MDR0_QUAD_MASK		~0x03


// 8-bit Mode Register 1
#define MDR1_MASK		~0xF7
// D7:		Flag on Carry.
//		= 0	NOP
//		= 1	FLAG on CY (D7 of STR)
#define MDR1_FLAG_CY		0x80
// D6:		Flag on Borrow.
//		= 0	NOP
//		= 1	FLAG on BW (D6 of STR)
#define MDR1_FLAG_BW		0x40
// D5:		Flag on Compare (CNTR == DTR).
//		= 0	NOP
//		= 1	FLAG on CMP (D5 of STR)
#define MDR1_FLAG_CMP		0x20
// D4:		Flag on Carry.
//		= 0	NOP
//		= 1	FLAG on IDX (D4 of STR)
#define MDR1_FLAG_IDX		0x10
#define MDR1_FLAG_MASK		~0xF0
// D3:		Unused
#define MDR1_UNUSED		0x08
// D2:		Enable counting.
// 		= 0	Enable
//		= 1	Disable
#define MDR1_COUNTER_ENABLE	0x00
#define MDR1_COUNTER_DISABLE	0x04
#define MDR1_COUNTER_MASK	~0x04
// D1 - D0:	Counter width.
//		= 00	4-byte
//		= 01	3-byte
//		= 10	2-byte
// 		= 11	1-byte
#define MDR1_WIDTH_4B		0x00
#define MDR1_WIDTH_3B		0x01
#define MDR1_WIDTH_2B		0x02
#define MDR1_WIDTH_1B		0x03
#define MDR1_WIDTH_MASK		~0x03


// 8-bit Status Register
#define STR_MASK		~0xFF
// D7:		Carry latch (CNTR overflow)
#define STR_CARRY		0x80
// D6:		Borrow latch (CTNR underflow)
#define STR_BORROW		0x40
// D5:		Compare latch (CNTR == DTR)
#define STR_COMPARE		0x20
// D4:		Index latch.
#define STR_INDEX		0x10
#define STR_LATCH_MASK		~0xF0
// D3:		Count enable status.
#define STR_COUNTER_ENABLED	0x08
// D2:		Power loss indicator latch
#define STR_POWERLOSS		0x04
// D1:		Count direction indicator
//		= 0	Down
//		= 1	Up
#define STR_DIRECTION		0x02
// D0:		Count sign bit
//		= 0	Positive
//		= 1	Negative
#define STR_SIGN		0x01

