// ----------------------------------------------------------------------------
/// readadc.c
/// © 2009-2012 Sealevel Systems, Inc.
/// <summary>
/// Sample demonstrating use of R9 Development board's analog inputs.
/// </summary>
///
/// <remarks>
/// This application acts as a simple analog input status display.  The act
/// of reading the analog inputs on the R9 development board is a multi-part
/// process consisting of two device interface interactions.  First one must
/// configure a multiplexor to route the proper pins to the ADC.  The 
/// configuration of the mux is accomplished through the same ARM gpio 
/// interface demonstrated in the Digtial_Ouput and Digital_Input samples
/// provided in this package.  The next step is to trigger the A/D conversion
/// by sending a command to the ADC chip.  This is done through a SPI 
/// transation.  Finally the value must be retreived through another SPI
/// transation.
/// 
/// Command line arguments are accepted to modify the conversion range.
/// The conversion range is selected by specifiying a value between 0 and 3
/// which correspond to [0, 5]V, [0, 10]V, [-5, 5]V, and [-10, 10]V respectively.
/// This tool simply displays the status of all R9 analog inputs.
/// </remarks>
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "gpio.h"
#include "r91001_map.h"

// ARM GPIO corresponding to the channel MUX is registered here
int register_mux()
{
	int i;

	// register any IO points we might need
	for (i = 0; i < ANALOG_INPUT_MUX_BITS + 1; i++)
	{
		if (ANALOG_INPUT_MUX[i] < 0)
		{
			continue;
		}

		if (gpio_register(ANALOG_INPUT_MUX[i]) != 0)
		{
			printf("Error registering multiplexor with sysfs!\n");
			return -1;
		}

		// Ensure points are configured as outputs
		if (gpio_set_direction(ANALOG_INPUT_MUX[i], "out") != 0)
		{
			printf("Error setting direction of multiplexor\n");
			return -2;
		}
	}

	return 0;
}

// ARM GPIO corresponding to the channel MUX is released here
void release_mux()
{
	int i;

	// if you no longer wish to access the points, you can unregister them
	for (i = 0; i < ANALOG_INPUT_MUX_BITS + 1; i++)
	{
		if (ANALOG_INPUT_MUX[i] < 0)
		{
			continue;
		}

		gpio_unregister(ANALOG_INPUT_MUX[i]);
	}
}

// Similar to accessing Digital Outputs, there is an ADC channel
// multiplexor which we have to configure to route the right pins
// to the ADC.
int set_mux(int channel)
{
	int i;

	if (channel < 0 || channel > ((1 << ANALOG_INPUT_MUX_BITS) - 1))
	{
		printf("A/D Channel %i out of range\n", channel);
		return -1;
	}

	// Disable the mux while changing channels
	if (ANALOG_INPUT_MUX[0] >= 0)
	{
		if (gpio_set_value(ANALOG_INPUT_MUX[0], 0) != 0)
		{
			printf("Unable to set multiplexor\n");
			return -2;
		}
	}

	// Convert the decimal channel number into a binary address value
	// and apply it to the mux
	for (i = 0; i < ANALOG_INPUT_MUX_BITS; i++)
	{
		int state = (channel >> i) & 0x01;
		
		if (gpio_set_value(ANALOG_INPUT_MUX[1 + i], state) != 0)
		{
			printf("Unable to set multiplexor\n");
			return -3;
		}
	}

	// Enable the mux now that we're done
	if (ANALOG_INPUT_MUX[0] >= 0)
	{
		if (gpio_set_value(ANALOG_INPUT_MUX[0], 1) != 0)
		{
			printf("Unable to set multiplexor\n");
			return -4;
		}
	}
	
	return 0;
}

int RawValue(unsigned char range, unsigned char *raw)
{
	// Convert the raw bytes into a signed or unsigned short
	// and adjust the value to be a 12-bit number (ADC determines this)
	if (range == ZERO_TO_FIVE || range == ZERO_TO_TEN)
	{
		unsigned short value = (raw[0] << 8) | (raw[1]);
		return (value / 16);
	}
	else
	{
		signed short value = (raw[0] << 8) | (raw[1]);
		return (value / 16);
	}
}

double Convert(unsigned char range, int raw)
{
	// Convert the digital ADC reading back into it's corresponding
	// analog voltage.
	int rise, run = 4095;
	double slope, offset = 0.0;
	
	switch (range)
	{
		case ZERO_TO_FIVE: 
			rise = 5;
			run = 4095;
			offset = 0.0;
			break;

		case ZERO_TO_TEN: 
			rise = 10;
			run = 4095;
			offset = 0.0;
			break;

		case MINUS_FIVE_TO_FIVE: 
			rise = 5; 
			run = (raw > 0) ? 2045 : 2048;
			offset = (raw > 0) ? (-2 * ((rise * 1.0) / run)) : 0.0;
			break;

		case MINUS_TEN_TO_TEN:
			rise = 10; 
			run = (raw > 0) ? 2046 : 2048;
			offset = (raw > 0) ? (-1 * ((rise * 1.0) / run)) : 0.0;
			break;

		default:
			printf("Error, unsupported range!\n"); 
			return 0.0;
	}

	slope = ((rise * 1.0) / run);

	return (raw * slope) + offset;
}

int main(int argc, char *args[])
{
	int i;
	int fd;
	unsigned char range = ZERO_TO_FIVE;

	// Any argument is expected to be a value [0,3] indicating
	// the conversion range to use
	if (argc > 1)
	{
		switch(atoi(args[1]))
		{
			default:
			case 0: range = ZERO_TO_FIVE; 		break;
			case 1:	range = ZERO_TO_TEN; 		break;
			case 2: range = MINUS_FIVE_TO_FIVE;	break;
			case 3: range = MINUS_TEN_TO_TEN;	break;
		}
	}

	fd = open(ANALOG_INPUT_DEV, O_RDWR, 0);
	if (fd < 0)
	{
		printf("Couldn't open spi device!\n");
		return -1;
	}

	// register the multiplexor
	register_mux();

	printf("AI  | Raw          | Volts\n");
	for (i = 0; i < (1 << ANALOG_INPUT_MUX_BITS); i++)
	{
		unsigned char command;
		unsigned char response[2];
		int rawValue;
		double value;

		// Route the correct pins to the ADC
		set_mux(i);

		// Request a conversion
		command = range;
		write(fd, &command, sizeof(command));
		
		// Read the response which should be the converted value
		read(fd, response, sizeof(response));

		rawValue = RawValue(range, response);
		value = Convert(range, rawValue);

		printf("%3i | %5i (%04X) | %f\n", (i + 1), rawValue, (unsigned short) rawValue, value);
	}
	
	// release the multiplexor if you are done
	release_mux();
	close(fd);

	return 0;
}
